/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozCmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#define GDK_NATIVE_WINDOW_POINTER
#include "SceneViewerSrc.h"
extern "C" {
#include "api/cmiss_scene_viewer.h"
#include "api/cmiss_texture.h"
}
#ifdef XP_UNIX
#include <gtk/gtk.h>
#endif
#ifdef XP_WIN32
#include <windows.h>
#include <stdio.h>
#endif
#include "CommandDataSrc.h"
#include "zincPlugin.h"
#include "zincCmguiSrc.h"

static CmguiSceneViewer *createSceneViewer(CmguiSceneViewerData *scene_viewer_data);

static int clearSceneViewerImpl(CmguiSceneViewer *scene_viewer);

#ifdef XP_WIN32
class CmguiSceneViewerData
{
public:
  Cmiss_scene_viewer_package *scene_viewer_package;
  zincIPlugin *plugin;
  CmguiCommandData *parent;
  CmguiISceneViewerResultMonitor *svrm;
  HWND childHWND;

  CmguiSceneViewerData(Cmiss_scene_viewer_package *scene_viewer_package,
     zincIPlugin *plugin, CmguiCommandData *parent, 
     CmguiISceneViewerResultMonitor *svrm) : 
    scene_viewer_package(scene_viewer_package),
    plugin(plugin), parent(parent), svrm(svrm)
  {
	  NS_ADDREF(svrm);
	  childHWND = (HWND)NULL;
  }
};
#endif
#ifdef XP_UNIX
class CmguiSceneViewerData
{
public:
  Cmiss_scene_viewer_package *scene_viewer_package;
  zincIPlugin *plugin;
  CmguiCommandData *parent;
  CmguiISceneViewerResultMonitor *svrm;
							     
  CmguiSceneViewerData(Cmiss_scene_viewer_package *scene_viewer_package,
     zincIPlugin *plugin, CmguiCommandData *parent,
     CmguiISceneViewerResultMonitor *svrm) :
    scene_viewer_package(scene_viewer_package),
    plugin(plugin), parent(parent), svrm(svrm)
  {
	  NS_ADDREF(svrm);
  }

	~CmguiSceneViewerData()
	{
		NS_RELEASE(svrm);
	}
};
#endif // XP_UNIX

class CmguiSceneViewerResizeMonitor : public zincIPluginResizeMonitor
{
  CmguiSceneViewer *scene_viewer;
  CmguiSceneViewerData *scene_viewer_data;

public:
  NS_DECL_ISUPPORTS
  NS_DECL_ZINCIPLUGINRESIZEMONITOR

  CmguiSceneViewerResizeMonitor(CmguiSceneViewerData *scene_viewer_data) :
    scene_viewer_data(scene_viewer_data)
  {
    /* Try to create the scene viewer straight away, if it fails then
     we will try again when the plugin calls us. */
    scene_viewer = createSceneViewer(scene_viewer_data);
#if defined (ZINC_DEBUG)
    printf("CmguiSceneViewerResizeMonitor:: Constructor: Called createSceneViewer : %p\n",
       scene_viewer);
#endif // defined (ZINC_DEBUG)
  }

protected:
  virtual ~CmguiSceneViewerResizeMonitor()
  {
  }

};
NS_IMPL_ISUPPORTS1(CmguiSceneViewerResizeMonitor, zincIPluginResizeMonitor)

	NS_IMETHODIMP CmguiSceneViewerResizeMonitor::OnResize(PRUint32 width, PRUint32 height,
		PRUint32 x, PRUint32 y)
	{
#if defined (ZINC_DEBUG)
		printf("CmguiSceneViewerResizeMonitor::OnResize: Resize called width %d height %d\n", width, height);
#endif // defined (ZINC_DEBUG)
		if (!scene_viewer)
		  {
#if defined (ZINC_DEBUG)
		    printf("CmguiSceneViewerResizeMonitor::OnResize: Calling createSceneViewer : %p\n", scene_viewer);
#endif // defined (ZINC_DEBUG)
		    /* Try to create the scene viewer straight away, if it fails then
		       we will try again when the plugin calls us. */
		    scene_viewer = createSceneViewer(scene_viewer_data);
#if defined (ZINC_DEBUG)
		    printf("CmguiSceneViewerResizeMonitor::OnResize: Called createSceneViewer : %p\n", scene_viewer);
#endif // defined (ZINC_DEBUG)
		  }
		if (scene_viewer)
		  {
			  scene_viewer->SetWindowSize(width, height, x, y);
		  }
		return (NS_OK);
	}

NS_IMETHODIMP CmguiSceneViewerResizeMonitor::OnClose()
	{
#if defined (ZINC_DEBUG)
		printf("CmguiSceneViewerResizeMonitor::OnClose: Close called\n");
#endif // defined (ZINC_DEBUG)
		if (scene_viewer)
		  {
		scene_viewer->Close();
		  }
		return (NS_OK);
	}

NS_IMETHODIMP CmguiSceneViewerResizeMonitor::OnEvent(void *NPevent)
	{
#if defined (ZINC_DEBUG)
		//printf("CmguiSceneViewerResizeMonitor::OnEvent: Called\n");
#endif // defined (ZINC_DEBUG)
#if defined (WIN32_USER_INTERFACE)
		if (scene_viewer)
		{
		  NPEvent *event = static_cast<NPEvent*>(NPevent);
		  Cmiss_scene_viewer_handle_windows_event(scene_viewer->impl,
		     event->event, event->wParam, event->lParam);
		}
#endif /* defined (WIN32_USER_INTERFACE) */
		return (NS_OK);
	}

#ifdef XP_UNIX
static void CmguiSceneViewer_gtk_unrealize_CB(GtkObject *object, gpointer scene_viewer_void)
/*******************************************************************************
LAST MODIFIED : 24 January 2007

DESCRIPTION :
Called when the gtk widget is being destroyed.  We need to clean up our reference
and if this is the last scene viewer destroy the command data while we still have
the correct OpenGL graphics context so that display lists are freed correctly.
(Alternatively we would have to not delete our display lists when destroying objects 
and rely on the graphics contexts to free them up instead).
==============================================================================*/
{
	CmguiSceneViewer *scene_viewer;

	if ((scene_viewer=(CmguiSceneViewer *)scene_viewer_void))
	{
#if defined (ZINC_DEBUG)
			printf("CmguiSceneViewer_gtk_unrealize_CB: Called.\n");
#endif // defined (ZINC_DEBUG)

		if (scene_viewer->impl)
		{
			// Make sure this scene_viewer is current
			Cmiss_scene_viewer_redraw_now(scene_viewer->impl);
			
#if defined (ZINC_DEBUG)
			printf("CmguiSceneViewer_gtk_unrealize_CB: Notifying CmguiCommandData\n");
#endif // defined (ZINC_DEBUG)
			
			if (scene_viewer->data->parent)
			{
				CmguiCommandData_closing_scene_viewer_notification(
					scene_viewer->data->parent, scene_viewer);
				scene_viewer->data->parent = (CmguiCommandData *)NULL;
			}
		}
		clearSceneViewerImpl(scene_viewer);
	}
	else
	{
		printf ("CmguiSceneViewer_gtk_unrealize_CB.  Missing scene_viewer\n");
	}
} /* CmguiSceneViewer_gtk_close_CB */
#endif // XP_UNIX

CmguiSceneViewer::CmguiSceneViewer()
{
	impl = (Cmiss_scene_viewer_id)NULL;
	data = (CmguiSceneViewerData*)NULL;
}

CmguiSceneViewer::~CmguiSceneViewer()
{
	if (data)
	{
		delete data;
	}
}

#ifdef XP_WIN32
static void process_scene_viewer_repaint_required(
	Cmiss_scene_viewer_id sceneViewer,
	void* callbackData, void* userData)
{
  // userData is cmguiSceneViewer...
	CmguiSceneViewer *sceneviewer =
    static_cast<CmguiSceneViewer*>(userData);

  zincIPlugin *plugin = sceneviewer->data->plugin;
  plugin->RepaintRequired();
}
#endif /* XP_WIN32 */

static CmguiSceneViewer *createSceneViewer(CmguiSceneViewerData *scene_viewer_data)
{
  zincIPlugin *plugin = scene_viewer_data->plugin;
  zincIPluginWindow *plugin_window = NULL;

  if ((NULL == plugin)
    || (NS_OK != plugin->QueryInterface(
    NS_GET_IID(zincIPluginWindow), (void **)&plugin_window))
    || (NULL == plugin_window))
  {
     return 0;
  }
  
#ifdef XP_WIN32
  void *window_handle;
  void *drawable_handle;

  if (NS_OK != plugin_window->GetWindow(&window_handle))
  {
    return NULL;
  }
  if (NS_OK != plugin_window->GetDrawable(&drawable_handle))
  {
    return NULL;
  }
  if (!window_handle && !drawable_handle)
  {
#if defined (ZINC_DEBUG)
	  printf("createSceneViewer: window or drawable not set, not creating now\n");
#endif // defined (ZINC_DEBUG)
	  return NULL;
  }
  
  HWND hWND = static_cast<HWND>(window_handle);
  HDC hDC = static_cast<HDC>(drawable_handle);

  PRUint32 width, height;

  if (NS_OK != plugin->GetWidth(&width))
  {
	  width = 100;
  }
  if (NS_OK != plugin->GetHeight(&height))
  {
	  height = 100;
  }
  
#if defined (ZINC_DEBUG)
  printf("createSceneViewer: width %d height %d window %p drawable %p\n",
     width, height, window_handle, drawable_handle);
#endif // defined (ZINC_DEBUG)

  HWND childHWND;

  if (hWND)
  {
	  /* We must stop the parent window provided by mozilla from drawing over our new child window
		  by adding WS_CLIPCHILDREN to its window style */
	  SetWindowLong(hWND, GWL_STYLE, WS_VISIBLE | WS_CHILDWINDOW | WS_CLIPCHILDREN);

	  WNDCLASS srvWC;

	  srvWC.style = CS_HREDRAW | CS_VREDRAW;
	  srvWC.lpfnWndProc = DefWindowProc;
	  srvWC.cbClsExtra = 0;
	  srvWC.cbWndExtra = 0;
	  srvWC.hInstance = GetModuleHandle(NULL);
	  srvWC.hIcon = NULL;
	  srvWC.hCursor = LoadCursor(NULL, IDC_ARROW);
	  srvWC.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	  srvWC.lpszMenuName = NULL;
	  srvWC.lpszClassName = "cmgui_embed_wndclass";
	  RegisterClass(&srvWC);

	  /* I am not sure that this child window is required any longer and 
		  if it is not performing any useful function could be removed. */
	  childHWND = CreateWindow(
		  "cmgui_embed_wndclass",
		  "",
		  WS_CHILD | WS_VISIBLE,
		  0, 0,
		  width, height,
		  hWND,
		  NULL,
		  GetModuleHandle(NULL),
		  NULL
		  );
	  if (childHWND == NULL)
	  {
		  printf("createSceneViewer: CreateWindow returned a NULL pointer.\n");
		  return NULL;
	  }
  }
  else
  {
	  childHWND = NULL;
  }

#if defined (ZINC_DEBUG)
  printf("createSceneViewer: scene_viewer_data %p scene_viewer_package %p svrm %p\n",
     scene_viewer_data, scene_viewer_data->scene_viewer_package,
	  scene_viewer_data->svrm);
#endif // defined (ZINC_DEBUG)

  Cmiss_scene_viewer_id impl = CREATE(Cmiss_scene_viewer_win32)
	  (scene_viewer_data->scene_viewer_package,
		  childHWND, hDC,
		  CMISS_SCENE_VIEWER_BUFFERING_DOUBLE,
		  CMISS_SCENE_VIEWER_STEREO_ANY_MODE,
		  /*minimum_colour_buffer_depth*/5,
		  /*minimum_depth_buffer_depth*/5, 
		  /*minimum_accumulation_buffer_depth*/0);
  if (impl == NULL)
  {
	  printf("CREATE(Cmiss_scene_viewer_win32) returned NULL.\n");
	  return NULL;
  }
  
  scene_viewer_data->childHWND = childHWND;

  CmguiSceneViewer *scene_viewer = new CmguiSceneViewer(impl, scene_viewer_data);

  if (!childHWND)
  {
	  Cmiss_scene_viewer_add_repaint_required_callback(impl,
		  process_scene_viewer_repaint_required, scene_viewer);  
  }

#elif defined XP_UNIX
	void *window_handle;
	
	if (NS_OK != plugin_window->GetWindow(&window_handle))
	{
		return NULL;
	}

	GdkNativeWindow xid = reinterpret_cast<GdkNativeWindow>(window_handle);
	GtkWidget *windowPlug, *packBox;

	windowPlug = gtk_plug_new(xid);
	if (windowPlug == NULL)
	{
      printf("gtk_plug_new(%p) returned NULL.\n", xid);
      return NULL;
	}

	packBox = gtk_hbox_new(FALSE, 0);
	if (packBox == NULL)
	{
      printf("gtk_hbox_new(FALSE, 0) returned NULL.\n");
      return NULL;
	}

	Cmiss_scene_viewer_id impl = create_Cmiss_scene_viewer_gtk
	  (scene_viewer_data->scene_viewer_package,
		  GTK_CONTAINER(packBox),
		  CMISS_SCENE_VIEWER_BUFFERING_DOUBLE,
		  CMISS_SCENE_VIEWER_STEREO_ANY_MODE,
		  /*minimum_colour_buffer_depth*/5,
		  /*minimum_depth_buffer_depth*/5, 
		  /*minimum_accumulation_buffer_depth*/0);

	if (impl == NULL)
	{
      printf("create_Cmiss_scene_viewer_gtk() returned NULL.\n");
      return NULL;
	}

	gtk_widget_show(packBox);
	gtk_container_add(GTK_CONTAINER(windowPlug), packBox);
	gtk_widget_show(windowPlug);

	CmguiSceneViewer *scene_viewer = new CmguiSceneViewer(impl,
		scene_viewer_data);

	// The destroy callback happens too late, the window is already gone so
	// we cannot make it current.  (However we are primarily relying on the
	// close notification from the plugin, it seems to be first.)
	g_signal_connect (G_OBJECT(packBox), "unrealize",
			G_CALLBACK(CmguiSceneViewer_gtk_unrealize_CB), (gpointer)scene_viewer);

#endif

	CmguiCommandData_opening_scene_viewer_notification(
		scene_viewer_data->parent, scene_viewer);

	NS_ADDREF(scene_viewer);
	scene_viewer_data->svrm->SceneViewerReady(scene_viewer);

	return(scene_viewer);
}

int clearSceneViewerImpl(CmguiSceneViewer *scene_viewer)
{
	if (scene_viewer->impl)
	{
#ifdef WIN32
#if defined (ZINC_DEBUG)
		printf("clearSceneViewerImpl: Setting the window prod to DefWindowProc\n");
#endif // defined (ZINC_DEBUG)
		SetWindowLong(scene_viewer->data->childHWND, 
			GWL_WNDPROC, (long)DefWindowProc);
#endif
		scene_viewer->impl = (Cmiss_scene_viewer*)NULL;
	}
	return(1);
}

/* void close (); */
NS_IMETHODIMP CmguiSceneViewer::Close()
{
	if (impl)
	{
		// Make sure this scene_viewer is current
		Cmiss_scene_viewer_redraw_now(impl);
		
#if defined (ZINC_DEBUG)
		printf("CmguiSceneViewer::Close: Notifying CmguiCommandData %p\n", impl);
#endif // defined (ZINC_DEBUG)
	

		if (data->parent)
		{
			CmguiCommandData_closing_scene_viewer_notification(
				data->parent, this);
			data->parent = (CmguiCommandData *)NULL;
		}
	}

	// Need to recheck to see if the command data has already cleared out the
	// implementation
	if (impl)
	{
#if defined (ZINC_DEBUG)
		printf("CmguiSceneViewer::Close: Calling Cmiss_scene_viewer_destroy %p\n", impl);
#endif // defined (ZINC_DEBUG)
		Cmiss_scene_viewer_destroy(&impl);
		clearSceneViewerImpl(this);
	}
	return NS_OK;
}

/* void setWindowSize (in unsigned long width, in unsigned long height); */
NS_IMETHODIMP CmguiSceneViewer::SetWindowSize(PRUint32 width, PRUint32 height,
	PRUint32 x, PRUint32 y)
{
#ifdef WIN32
	if (data->childHWND)
	{
		SetWindowPos(data->childHWND, HWND_TOP, 0, 0, width, height,
			SWP_NOZORDER | SWP_NOMOVE);
		SendMessage(data->childHWND, WM_SIZING, 0, 0);
	}
	else
	{
		Cmiss_scene_viewer_win32_set_window_size(impl, width, height,
			x, y);
	}
#else
	//width = 0;
	//height = 0;
  /* XXX fixme */
#endif
  return NS_OK;
}

int createSceneViewerFromPlugin(
   Cmiss_scene_viewer_package *scene_viewer_package, zincIPlugin *plugin,
   CmguiCommandData *parent, CmguiISceneViewerResultMonitor *svrm)
{
  /* We need to create the resize monitor now but we can only make the
     CmguiSceneViewer if the window handle has been set in the plugin already */

  //Need a resize monitor
  plugin->SetResizeMonitor(new CmguiSceneViewerResizeMonitor(
     new CmguiSceneViewerData(scene_viewer_package, plugin, parent,
     svrm)));

  return(1);
}

NS_IMETHODIMP CmguiSceneViewer::SetBackgroundTexturePosition(double left, double top, double width, double height)
{
  if (!Cmiss_scene_viewer_set_background_texture_info(impl,
        left, top, width, height, /*Don't undistort*/0,
		  /*Something above zero*/100))
  {
	  return NS_ERROR_UNEXPECTED;
  }
  return NS_OK;
}

NS_IMETHODIMP CmguiSceneViewer::GetFreespinTumbleAxis(float *axisX, float *axisY, float *axisZ)
{
	float axis[3];
	if (!Cmiss_scene_viewer_get_freespin_tumble_axis(impl, axis))
	{
		return NS_ERROR_UNEXPECTED;
	}
	*axisX = axis[0];
	*axisY = axis[1];
	*axisZ = axis[2];
	return NS_OK;
}

NS_IMETHODIMP CmguiSceneViewer::StartFreespin(float tumbleAxisX, float tumbleAxisY, float tumbleAxisZ, double tumbleAngle)
{
	float axis[3] = {tumbleAxisX, tumbleAxisY, tumbleAxisZ};
	if (!Cmiss_scene_viewer_start_freespin(impl, axis, tumbleAngle))
	{
		return NS_ERROR_UNEXPECTED;
	}
	return NS_OK;
}

void CmguiSceneViewer_command_data_destroy_notification(
	CmguiSceneViewer *cmguiSceneViewer, CmguiCommandData *cmguiCommandData)
{
	//Scene viewer has been destroyed by the command data
	clearSceneViewerImpl(cmguiSceneViewer);
}

NS_IMETHODIMP CmguiSceneViewer::GetFramePixels(PRUint32 storage, PRUint32 *width, PRUint32 *height, PRUint32 preferred_antialias, PRUint32 preferred_transparency_layers, void * *frame_data, PRUint32 force_onscreen)
{
  if (!(Cmiss_scene_viewer_get_frame_pixels(impl, (Cmiss_texture_storage_type)storage, (int *)width, (int *)height, (int)preferred_antialias, (int)preferred_transparency_layers, (unsigned char **) frame_data, (int)force_onscreen)))
  {
    return NS_ERROR_UNEXPECTED;
  }
  return NS_OK;
}

static void process_scene_viewer_transform(
	Cmiss_scene_viewer_id sceneViewer,
	void* callbackData, void* userData)
{
  // userData is an event handler...
	CmguiIEventHandler *event_handler =
    static_cast<CmguiIEventHandler*>(userData);

  event_handler->OnEvent();
}

/* [cmguihandcoded] void addTransformCallback (in CmguiIEventHandler handler); */
NS_IMETHODIMP CmguiSceneViewer::AddTransformCallback(CmguiIEventHandler *handler)
{
	if (!Cmiss_scene_viewer_add_transform_callback(impl,
			process_scene_viewer_transform, static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	NS_ADDREF(handler);
	return NS_OK;
}

/* [cmguihandcoded] void removeTransformCallback (in CmguiIEventHandler handler); */
NS_IMETHODIMP CmguiSceneViewer::RemoveTransformCallback(CmguiIEventHandler *handler)
{
	if (!Cmiss_scene_viewer_remove_transform_callback(impl,
			process_scene_viewer_transform, static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	NS_RELEASE(handler);
	return NS_OK;
}

static int process_scene_viewer_input(
	Cmiss_scene_viewer_id sceneViewer,
	Cmiss_scene_viewer_input_id sceneViewerInput, void* userData)
{
	int return_code;

	//Wrap the input
	CmguiISceneViewerInput *input = new CmguiSceneViewerInput(sceneViewerInput);
	NS_ADDREF(input);

  // userData is an event handler...
	CmguiISceneViewerInputHandler *event_handler =
    static_cast<CmguiISceneViewerInputHandler*>(userData);

	PRBool retval;
	event_handler->OnInput(input, &retval);
	if (retval)
	{
		return_code = 1;
	}
	else
	{
		return_code = 0;
	}

	NS_RELEASE(input);
	return(return_code);
}

/* [cmguihandcoded] void addInputCallback (in CmguiISceneViewerInputHandler handler, in boolean add_first); */
NS_IMETHODIMP CmguiSceneViewer::AddInputCallback(CmguiISceneViewerInputHandler *handler, PRBool add_first)
{
	if (!Cmiss_scene_viewer_add_input_callback(impl,
			process_scene_viewer_input, static_cast<void*>(handler), add_first))
	{
      return NS_ERROR_UNEXPECTED;
	}
	NS_ADDREF(handler);
	return NS_OK;
}

/* [cmguihandcoded] void removeInputCallback (in CmguiISceneViewerInputHandler handler); */
NS_IMETHODIMP CmguiSceneViewer::RemoveInputCallback(CmguiISceneViewerInputHandler *handler)
{
	if (!Cmiss_scene_viewer_remove_input_callback(impl,
			process_scene_viewer_input, static_cast<void*>(handler)))
	{
      return NS_ERROR_UNEXPECTED;
	}
	NS_RELEASE(handler);
	return NS_OK;
}


CmguiSceneViewerInput::CmguiSceneViewerInput()
{
	impl = (Cmiss_scene_viewer_input_id)NULL;
	data = (CmguiSceneViewerInputData*)NULL;
}

CmguiSceneViewerInput::~CmguiSceneViewerInput()
{
}

/* const unsigned long MODIFIER_SHIFT = 1; */
/* const unsigned long MODIFIER_CONTROL = 2; */
/* const unsigned long MODIFIER_ALT = 4; */
/* const unsigned long MODIFIER_BUTTON1 = 8; */
/* attribute unsigned long modifierFlags; */
NS_IMETHODIMP CmguiSceneViewerInput::GetModifierFlags(PRUint32 *aModifierFlags)
{
   enum Cmiss_scene_viewer_input_modifier_flags modifier_flags;

   if (!Cmiss_scene_viewer_input_get_modifier_flags(impl,
      &modifier_flags))
   {
     return NS_ERROR_UNEXPECTED;
   }
	*aModifierFlags = 0;
	if (modifier_flags & CMISS_SCENE_VIEWER_INPUT_MODIFIER_SHIFT)
	{
		*aModifierFlags |= MODIFIER_SHIFT;
	}
	if (modifier_flags & CMISS_SCENE_VIEWER_INPUT_MODIFIER_CONTROL)
	{
		*aModifierFlags |= MODIFIER_CONTROL;
	}
	if (modifier_flags & CMISS_SCENE_VIEWER_INPUT_MODIFIER_ALT)
	{
		*aModifierFlags |= MODIFIER_ALT;
	}
	if (modifier_flags & CMISS_SCENE_VIEWER_INPUT_MODIFIER_BUTTON1)
	{
		*aModifierFlags |= MODIFIER_BUTTON1;
	}
   return NS_OK;
}

NS_IMETHODIMP CmguiSceneViewerInput::SetModifierFlags(PRUint32 aModifierFlags)
{
   int modifier_flags, return_code;
	
	modifier_flags = 0;
	if (aModifierFlags & MODIFIER_SHIFT)
	{
		modifier_flags |= CMISS_SCENE_VIEWER_INPUT_MODIFIER_SHIFT;
	}
	if (aModifierFlags & MODIFIER_CONTROL)
	{
 		modifier_flags |= CMISS_SCENE_VIEWER_INPUT_MODIFIER_CONTROL;
	}
	if (aModifierFlags & MODIFIER_ALT)
	{
 		modifier_flags |= CMISS_SCENE_VIEWER_INPUT_MODIFIER_ALT;
	}
	if (aModifierFlags & MODIFIER_BUTTON1)
	{
 		modifier_flags |= CMISS_SCENE_VIEWER_INPUT_MODIFIER_BUTTON1;
	}
   if (!Cmiss_scene_viewer_input_set_modifier_flags(impl,
		static_cast<enum Cmiss_scene_viewer_input_modifier_flags>
		(modifier_flags)))
   {
     return NS_ERROR_UNEXPECTED;
   }
	return NS_OK;
}

